@@ -42,15 +42,27 @@ extern "C" {
 #endif
 
 #if defined(_COMPILER_ASAN_ENABLED_)
-static inline void sanitizer_start_switch_fiber(const void* bottom, size_t size) {
-    __sanitizer_start_switch_fiber(NULL, bottom, size);
+static inline void sanitizer_start_switch_fiber(jl_ptls_t ptls, jl_task_t *from, jl_task_t *to) {
+    if (to->copy_stack)
+        __sanitizer_start_switch_fiber(&from->ctx.asan_fake_stack, (char*)ptls->stackbase-ptls->stacksize, ptls->stacksize);
+    else
+        __sanitizer_start_switch_fiber(&from->ctx.asan_fake_stack, to->stkbuf, to->bufsz);
 }
-static inline void sanitizer_finish_switch_fiber(void) {
-    __sanitizer_finish_switch_fiber(NULL, NULL, NULL);
+static inline void sanitizer_start_switch_fiber_killed(jl_ptls_t ptls, jl_task_t *to) {
+    if (to->copy_stack)
+        __sanitizer_start_switch_fiber(NULL, (char*)ptls->stackbase-ptls->stacksize, ptls->stacksize);
+    else
+        __sanitizer_start_switch_fiber(NULL, to->stkbuf, to->bufsz);
+}
+static inline void sanitizer_finish_switch_fiber(jl_task_t *last, jl_task_t *current) {
+    __sanitizer_finish_switch_fiber(current->ctx.asan_fake_stack, NULL, NULL);
+        //(const void**)&last->stkbuf,
+        //&last->bufsz);
 }
 #else
-static inline void sanitizer_start_switch_fiber(const void* bottom, size_t size) {}
-static inline void sanitizer_finish_switch_fiber(void) {}
+static inline void sanitizer_start_switch_fiber(jl_ptls_t ptls, jl_task_t *from, jl_task_t *to) JL_NOTSAFEPOINT {}
+static inline void sanitizer_start_switch_fiber_killed(jl_ptls_t ptls, jl_task_t *to) JL_NOTSAFEPOINT {}
+static inline void sanitizer_finish_switch_fiber(jl_task_t *last, jl_task_t *current) JL_NOTSAFEPOINT {}
 #endif
 
 #if defined(_COMPILER_TSAN_ENABLED_)
@@ -109,7 +121,11 @@ static inline void sanitizer_finish_switch_fiber(void) {}
 #define MINSTKSZ 131072
 #endif
 
+#ifdef _COMPILER_ASAN_ENABLED_
+#define ROOT_TASK_STACK_ADJUSTMENT 0
+#else
 #define ROOT_TASK_STACK_ADJUSTMENT 3000000
+#endif
 
 #ifdef JL_HAVE_ASYNCIFY
 // Switching logic is implemented in JavaScript
@@ -133,13 +149,62 @@ static int always_copy_stacks = 1;
 static int always_copy_stacks = 0;
 #endif
 
+#if defined(_COMPILER_ASAN_ENABLED_)
+extern void __asan_get_shadow_mapping(size_t *shadow_scale, size_t *shadow_offset);
+
+JL_NO_ASAN void *memcpy_noasan(void *dest, const void *src, size_t n) {
+  char *d = (char*)dest;
+  const char *s = (const char *)src;
+  for (size_t i = 0; i < n; ++i)
+    d[i] = s[i];
+  return dest;
+}
+
+JL_NO_ASAN void *memcpy_a16_noasan(uint64_t *dest, const uint64_t *src, size_t nb) {
+  uint64_t *end = (uint64_t*)((char*)src + nb);
+  while (src < end)
+    *(dest++) = *(src++);
+  return dest;
+}
+
+/* Copy stack are allocated as regular bigval objects and do no go through free_stack,
+   which would otherwise unpoison it before returning to the GC pool */
+static void asan_free_copy_stack(void *stkbuf, size_t bufsz) {
+    __asan_unpoison_stack_memory((uintptr_t)stkbuf, bufsz);
+}
+#else
+static void asan_free_copy_stack(void *stkbuf, size_t bufsz) {}
+#endif
+
 #ifdef COPY_STACKS
-static void memcpy_a16(uint64_t *to, uint64_t *from, size_t nb)
+static void JL_NO_ASAN JL_NO_MSAN memcpy_stack_a16(uint64_t *to, uint64_t *from, size_t nb)
 {
+#if defined(_COMPILER_ASAN_ENABLED_)
+    /* Asan keeps shadow memory for everything on the stack. However, in general,
+       this function may touch invalid portions of the stack, since it just moves
+       the stack around. To keep ASAN's stack tracking capability intact, we need
+       to move the shadow memory along with the stack memory itself. */
+    size_t shadow_offset;
+    size_t shadow_scale;
+    __asan_get_shadow_mapping(&shadow_scale, &shadow_offset);
+    uintptr_t from_addr = (((uintptr_t)from) >> shadow_scale) + shadow_offset;
+    uintptr_t to_addr = (((uintptr_t)to) >> shadow_scale) + shadow_offset;
+    // Make sure that the shadow scale is compatible with the alignment, so
+    // we can copy whole bytes.
+    assert(shadow_scale <= 4);
+    size_t shadow_nb = nb >> shadow_scale;
+    // Copy over the shadow memory
+    memcpy_noasan((char*)to_addr, (char*)from_addr, shadow_nb);
+    memcpy_a16_noasan(jl_assume_aligned(to, 16), jl_assume_aligned(from, 16), nb);
+#elif defined(_COMPILER_MSAN_ENABLED_)
+# warning This function is imcompletely implemented for MSAN (TODO).
+    memcpy((char*)jl_assume_aligned(to, 16), (char*)jl_assume_aligned(from, 16), nb);
+#else
     memcpy((char*)jl_assume_aligned(to, 16), (char*)jl_assume_aligned(from, 16), nb);
     //uint64_t *end = (uint64_t*)((char*)from + nb);
     //while (from < end)
     //    *(to++) = *(from++);
+#endif
 }
 
 static void NOINLINE save_stack(jl_ptls_t ptls, jl_task_t *lastt, jl_task_t **pt)
@@ -150,6 +215,7 @@ static void NOINLINE save_stack(jl_ptls_t ptls, jl_task_t *lastt, jl_task_t **pt
     size_t nb = stackbase - frame_addr;
     void *buf;
     if (lastt->bufsz < nb) {
+        asan_free_copy_stack(lastt->stkbuf, lastt->bufsz);
         buf = (void*)jl_gc_alloc_buf(ptls, nb);
         lastt->stkbuf = buf;
         lastt->bufsz = nb;
@@ -160,14 +226,14 @@ static void NOINLINE save_stack(jl_ptls_t ptls, jl_task_t *lastt, jl_task_t **pt
     *pt = NULL; // clear the gc-root for the target task before copying the stack for saving
     lastt->copy_stack = nb;
     lastt->sticky = 1;
-    memcpy_a16((uint64_t*)buf, (uint64_t*)frame_addr, nb);
+    memcpy_stack_a16((uint64_t*)buf, (uint64_t*)frame_addr, nb);
     // this task's stack could have been modified after
     // it was marked by an incremental collection
     // move the barrier back instead of walking it again here
     jl_gc_wb_back(lastt);
 }
 
-static void NOINLINE JL_NORETURN restore_stack(jl_task_t *t, jl_ptls_t ptls, char *p)
+JL_NO_ASAN static void NOINLINE JL_NORETURN restore_stack(jl_task_t *t, jl_ptls_t ptls, char *p)
 {
     size_t nb = t->copy_stack;
     char *_x = (char*)ptls->stackbase - nb;
@@ -181,9 +247,8 @@ static void NOINLINE JL_NORETURN restore_stack(jl_task_t *t, jl_ptls_t ptls, cha
     }
     void *_y = t->stkbuf;
     assert(_x != NULL && _y != NULL);
-    memcpy_a16((uint64_t*)_x, (uint64_t*)_y, nb); // destroys all but the current stackframe
+    memcpy_stack_a16((uint64_t*)_x, (uint64_t*)_y, nb); // destroys all but the current stackframe
 
-    sanitizer_start_switch_fiber(t->stkbuf, t->bufsz);
 #if defined(_OS_WINDOWS_)
     jl_setcontext(&t->ctx.copy_ctx);
 #else
@@ -192,14 +257,14 @@ static void NOINLINE JL_NORETURN restore_stack(jl_task_t *t, jl_ptls_t ptls, cha
     abort(); // unreachable
 }
 
-static void restore_stack2(jl_task_t *t, jl_ptls_t ptls, jl_task_t *lastt)
+JL_NO_ASAN static void restore_stack2(jl_task_t *t, jl_ptls_t ptls, jl_task_t *lastt)
 {
     assert(t->copy_stack && !lastt->copy_stack);
     size_t nb = t->copy_stack;
     char *_x = (char*)ptls->stackbase - nb;
     void *_y = t->stkbuf;
     assert(_x != NULL && _y != NULL);
-    memcpy_a16((uint64_t*)_x, (uint64_t*)_y, nb); // destroys all but the current stackframe
+    memcpy_stack_a16((uint64_t*)_x, (uint64_t*)_y, nb); // destroys all but the current stackframe
 #if defined(JL_HAVE_UNW_CONTEXT)
     volatile int returns = 0;
     int r = unw_getcontext(&lastt->ctx.ctx);
@@ -213,7 +278,6 @@ static void restore_stack2(jl_task_t *t, jl_ptls_t ptls, jl_task_t *lastt)
 #else
 #error COPY_STACKS is incompatible with this platform
 #endif
-    sanitizer_start_switch_fiber(t->stkbuf, t->bufsz);
     tsan_switch_to_copyctx(&t->ctx);
 #if defined(_OS_WINDOWS_)
     jl_setcontext(&t->ctx.copy_ctx);
@@ -235,8 +299,10 @@ void JL_NORETURN jl_finish_task(jl_task_t *t)
         jl_atomic_store_release(&t->_state, JL_TASK_STATE_FAILED);
     else
         jl_atomic_store_release(&t->_state, JL_TASK_STATE_DONE);
-    if (t->copy_stack) // early free of stkbuf
+    if (t->copy_stack) { // early free of stkbuf
+        asan_free_copy_stack(t->stkbuf, t->bufsz);
         t->stkbuf = NULL;
+    }
     // ensure that state is cleared
     ct->ptls->in_finalizer = 0;
     ct->ptls->in_pure_callback = 0;
@@ -254,7 +320,7 @@ void JL_NORETURN jl_finish_task(jl_task_t *t)
             jl_apply(args, 2);
         }
         JL_CATCH {
-            jl_no_exc_handler(jl_current_exception());
+            jl_no_exc_handler(jl_current_exception(), ct);
         }
     }
     jl_gc_debug_critical_error();
@@ -265,7 +331,8 @@ JL_DLLEXPORT void *jl_task_stack_buffer(jl_task_t *task, size_t *size, int *ptid
 {
     size_t off = 0;
 #ifndef _OS_WINDOWS_
-    if (jl_all_tls_states[0]->root_task == task) {
+    jl_ptls_t ptls0 = jl_atomic_load_relaxed(&jl_all_tls_states)[0];
+    if (ptls0->root_task == task) {
         // See jl_init_root_task(). The root task of the main thread
         // has its buffer enlarged by an artificial 3000000 bytes, but
         // that means that the start of the buffer usually points to
@@ -306,7 +373,8 @@ JL_DLLEXPORT void jl_active_task_stack(jl_task_t *task,
     else if (task->stkbuf) {
         *total_start = *active_start = (char*)task->stkbuf;
 #ifndef _OS_WINDOWS_
-        if (jl_all_tls_states[0]->root_task == task) {
+        jl_ptls_t ptls0 = jl_atomic_load_relaxed(&jl_all_tls_states)[0];
+        if (ptls0->root_task == task) {
             // See jl_init_root_task(). The root task of the main thread
             // has its buffer enlarged by an artificial 3000000 bytes, but
             // that means that the start of the buffer usually points to
@@ -362,7 +430,7 @@ JL_DLLEXPORT jl_task_t *jl_get_next_task(void) JL_NOTSAFEPOINT
 const char tsan_state_corruption[] = "TSAN state corrupted. Exiting HARD!\n";
 #endif
 
-static void ctx_switch(jl_task_t *lastt)
+JL_NO_ASAN static void ctx_switch(jl_task_t *lastt)
 {
     jl_ptls_t ptls = lastt->ptls;
     jl_task_t **pt = &ptls->next_task;
@@ -416,7 +484,7 @@ static void ctx_switch(jl_task_t *lastt)
         if (lastt->copy_stack) { // save the old copy-stack
             save_stack(ptls, lastt, pt); // allocates (gc-safepoint, and can also fail)
             if (jl_setjmp(lastt->ctx.copy_ctx.uc_mcontext, 0)) {
-                sanitizer_finish_switch_fiber();
+                sanitizer_finish_switch_fiber(ptls->previous_task, jl_atomic_load_relaxed(&ptls->current_task));
                 // TODO: mutex unlock the thread we just switched from
                 return;
             }
@@ -441,12 +509,25 @@ static void ctx_switch(jl_task_t *lastt)
     if (t->started) {
 #ifdef COPY_STACKS
         if (t->copy_stack) {
-            if (!killed && !lastt->copy_stack)
+            if (lastt->copy_stack) {
+                // Switching from copystack to copystack. Clear any shadow stack
+                // memory above the saved shadow stack.
+                uintptr_t stacktop = (uintptr_t)ptls->stackbase - t->copy_stack;
+                uintptr_t stackbottom = ((uintptr_t)jl_get_frame_addr() & ~15);
+                if (stackbottom < stacktop)
+                    asan_unpoison_stack_memory(stackbottom, stacktop-stackbottom);
+            }
+            if (!killed && !lastt->copy_stack) {
+                sanitizer_start_switch_fiber(ptls, lastt, t);
                 restore_stack2(t, ptls, lastt);
-            else {
+            } else {
                 tsan_switch_to_copyctx(&t->ctx);
-                if (killed)
+                if (killed) {
+                    sanitizer_start_switch_fiber_killed(ptls, t);
                     tsan_destroy_copyctx(ptls, &lastt->ctx);
+                } else {
+                    sanitizer_start_switch_fiber(ptls, lastt, t);
+                }
 
                 if (lastt->copy_stack) {
                     restore_stack(t, ptls, NULL); // (doesn't return)
@@ -459,14 +540,26 @@ static void ctx_switch(jl_task_t *lastt)
         else
 #endif
         {
-            sanitizer_start_switch_fiber(t->stkbuf, t->bufsz);
+            if (lastt->copy_stack) {
+                // Switching away from a copystack to a non-copystack. Clear
+                // the whole shadow stack now, because otherwise we won't know
+                // how much stack memory to clear the next time we switch to
+                // a copystack.
+                uintptr_t stacktop = (uintptr_t)ptls->stackbase;
+                uintptr_t stackbottom = ((uintptr_t)jl_get_frame_addr() & ~15);
+                // We're not restoring the stack, but we still need to unpoison the
+                // stack, so it starts with a pristine stack.
+                asan_unpoison_stack_memory(stackbottom, stacktop-stackbottom);
+            }
             if (killed) {
+                sanitizer_start_switch_fiber_killed(ptls, t);
                 tsan_switch_to_ctx(&t->ctx);
                 tsan_destroy_ctx(ptls, &lastt->ctx);
                 jl_set_fiber(&t->ctx); // (doesn't return)
                 abort(); // unreachable
             }
             else {
+                sanitizer_start_switch_fiber(ptls, lastt, t);
                 if (lastt->copy_stack) {
                     // Resume at the jl_setjmp earlier in this function,
                     // don't do a full task swap
@@ -480,11 +573,20 @@ static void ctx_switch(jl_task_t *lastt)
         }
     }
     else {
-        sanitizer_start_switch_fiber(t->stkbuf, t->bufsz);
+        if (lastt->copy_stack) {
+            uintptr_t stacktop = (uintptr_t)ptls->stackbase;
+            uintptr_t stackbottom = ((uintptr_t)jl_get_frame_addr() & ~15);
+            // We're not restoring the stack, but we still need to unpoison the
+            // stack, so it starts with a pristine stack.
+            asan_unpoison_stack_memory(stackbottom, stacktop-stackbottom);
+        }
         if (t->copy_stack && always_copy_stacks) {
             tsan_switch_to_ctx(&t->ctx);
             if (killed) {
+                sanitizer_start_switch_fiber_killed(ptls, t);
                 tsan_destroy_ctx(ptls, &lastt->ctx);
+            } else {
+                sanitizer_start_switch_fiber(ptls, lastt, t);
             }
 #ifdef COPY_STACKS
 #if defined(_OS_WINDOWS_)
@@ -497,12 +599,14 @@ static void ctx_switch(jl_task_t *lastt)
         }
         else {
             if (killed) {
+                sanitizer_start_switch_fiber_killed(ptls, t);
                 tsan_switch_to_ctx(&t->ctx);
                 tsan_destroy_ctx(ptls, &lastt->ctx);
                 jl_start_fiber_set(&t->ctx); // (doesn't return)
                 abort();
             }
-            else if (lastt->copy_stack) {
+            sanitizer_start_switch_fiber(ptls, lastt, t);
+            if (lastt->copy_stack) {
                 // Resume at the jl_setjmp earlier in this function
                 tsan_switch_to_ctx(&t->ctx);
                 jl_start_fiber_set(&t->ctx); // (doesn't return)
@@ -513,7 +617,7 @@ static void ctx_switch(jl_task_t *lastt)
             }
         }
     }
-    sanitizer_finish_switch_fiber();
+    sanitizer_finish_switch_fiber(ptls->previous_task, jl_atomic_load_relaxed(&ptls->current_task));
 }
 
 JL_DLLEXPORT void jl_switch(void)
@@ -592,7 +696,7 @@ JL_DLLEXPORT void jl_switchto(jl_task_t **pt)
     jl_switch();
 }
 
-JL_DLLEXPORT JL_NORETURN void jl_no_exc_handler(jl_value_t *e)
+JL_DLLEXPORT JL_NORETURN void jl_no_exc_handler(jl_value_t *e, jl_task_t *ct)
 {
     // NULL exception objects are used when rethrowing. we don't have a handler to process
     // the exception stack, so at least report the exception at the top of the stack.
@@ -603,54 +707,83 @@ JL_DLLEXPORT JL_NORETURN void jl_no_exc_handler(jl_value_t *e)
     jl_static_show((JL_STREAM*)STDERR_FILENO, e);
     jl_printf((JL_STREAM*)STDERR_FILENO, "\n");
     jlbacktrace(); // written to STDERR_FILENO
+    if (ct == NULL)
+        jl_raise(6);
     jl_exit(1);
 }
 
-// yield to exception handler
+/* throw_internal - yield to exception handler */
+
+#ifdef ENABLE_TIMINGS
+#define pop_timings_stack()                                                    \
+        jl_timing_block_t *cur_block = ptls->timing_stack;                     \
+        while (cur_block && eh->timing_stack != cur_block) {                   \
+            cur_block = jl_pop_timing_block(cur_block);                        \
+        }                                                                      \
+        assert(cur_block == eh->timing_stack);
+#else
+#define pop_timings_stack() /* Nothing */
+#endif
+
+#define throw_internal_body()                                                  \
+    assert(!jl_get_safe_restore());                                            \
+    jl_ptls_t ptls = ct->ptls;                                                 \
+    ptls->io_wait = 0;                                                         \
+    JL_GC_PUSH1(&exception);                                                   \
+    jl_gc_unsafe_enter(ptls);                                                  \
+    if (exception) {                                                           \
+        /* The temporary ptls->bt_data is rooted by special purpose code in the\
+           GC. This exists only for the purpose of preserving bt_data until we \
+           set ptls->bt_size=0 below. */                                       \
+        jl_push_excstack(&ct->excstack, exception,                             \
+                          ptls->bt_data, ptls->bt_size);                       \
+        ptls->bt_size = 0;                                                     \
+    }                                                                          \
+    assert(ct->excstack && ct->excstack->top);                                 \
+    jl_handler_t *eh = ct->eh;                                                 \
+    if (eh != NULL) {                                                          \
+        pop_timings_stack()                                                    \
+        asan_unpoison_task_stack(ct, &eh->eh_ctx);                             \
+        jl_longjmp(eh->eh_ctx, 1);                                             \
+    }                                                                          \
+    else {                                                                     \
+        jl_no_exc_handler(exception, ct);                                      \
+    }                                                                          \
+    assert(0);
+
 static void JL_NORETURN throw_internal(jl_task_t *ct, jl_value_t *exception JL_MAYBE_UNROOTED)
 {
-    assert(!jl_get_safe_restore());
-    jl_ptls_t ptls = ct->ptls;
-    ptls->io_wait = 0;
-    JL_GC_PUSH1(&exception);
-    jl_gc_unsafe_enter(ptls);
-    if (exception) {
-        // The temporary ptls->bt_data is rooted by special purpose code in the
-        // GC. This exists only for the purpose of preserving bt_data until we
-        // set ptls->bt_size=0 below.
-        jl_push_excstack(&ct->excstack, exception,
-                          ptls->bt_data, ptls->bt_size);
-        ptls->bt_size = 0;
-    }
-    assert(ct->excstack && ct->excstack->top);
-    jl_handler_t *eh = ct->eh;
-    if (eh != NULL) {
-#ifdef ENABLE_TIMINGS
-        jl_timing_block_t *cur_block = ptls->timing_stack;
-        while (cur_block && eh->timing_stack != cur_block) {
-            cur_block = jl_pop_timing_block(cur_block);
-        }
-        assert(cur_block == eh->timing_stack);
-#endif
-        jl_longjmp(eh->eh_ctx, 1);
-    }
-    else {
-        jl_no_exc_handler(exception);
-    }
-    assert(0);
+CFI_NORETURN
+    throw_internal_body()
+    jl_unreachable();
 }
 
+#ifdef _COMPILER_ASAN_ENABLED_
+/* On the signal stack, we don't want to create any asan frames, but we do on the
+   normal, stack, so we split this function in two, depending on which context
+   we're calling it in */
+JL_NO_ASAN static void JL_NORETURN throw_internal_altstack(jl_task_t *ct, jl_value_t *exception JL_MAYBE_UNROOTED)
+{
+CFI_NORETURN
+    throw_internal_body()
+    jl_unreachable();
+}
+#else
+#define throw_internal_altstack throw_internal
+#endif
+
 // record backtrace and raise an error
 JL_DLLEXPORT void jl_throw(jl_value_t *e JL_MAYBE_UNROOTED)
 {
     assert(e != NULL);
     jl_jmp_buf *safe_restore = jl_get_safe_restore();
-    if (safe_restore)
-        jl_longjmp(*safe_restore, 1);
     jl_task_t *ct = jl_get_current_task();
-    if (ct == NULL) // During startup
-        jl_no_exc_handler(e);
-    JL_GC_PROMISE_ROOTED(ct);
+    if (safe_restore) {
+        asan_unpoison_task_stack(ct, safe_restore);
+        jl_longjmp(*safe_restore, 1);
+    }
+    if (ct == NULL) // During startup, or on other threads
+        jl_no_exc_handler(e, ct);
     record_backtrace(ct->ptls, 1);
     throw_internal(ct, e);
 }
@@ -668,17 +801,19 @@ JL_DLLEXPORT void jl_rethrow(void)
 // Special case throw for errors detected inside signal handlers.  This is not
 // (cannot be) called directly in the signal handler itself, but is returned to
 // after the signal handler exits.
-JL_DLLEXPORT void JL_NORETURN jl_sig_throw(void)
+JL_DLLEXPORT JL_NO_ASAN void JL_NORETURN jl_sig_throw(void)
 {
 CFI_NORETURN
     jl_jmp_buf *safe_restore = jl_get_safe_restore();
-    if (safe_restore)
-        jl_longjmp(*safe_restore, 1);
     jl_task_t *ct = jl_current_task;
+    if (safe_restore) {
+        asan_unpoison_task_stack(ct, safe_restore);
+        jl_longjmp(*safe_restore, 1);
+    }
     jl_ptls_t ptls = ct->ptls;
     jl_value_t *e = ptls->sig_exception;
     ptls->sig_exception = NULL;
-    throw_internal(ct, e);
+    throw_internal_altstack(ct, e);
 }
 
 JL_DLLEXPORT void jl_rethrow_other(jl_value_t *e JL_MAYBE_UNROOTED)
@@ -704,7 +839,7 @@ JL_DLLEXPORT void jl_rethrow_other(jl_value_t *e JL_MAYBE_UNROOTED)
    ACM Trans. Math. Softw., 2021.
 
    There is a pure Julia implementation in stdlib that tends to be faster when used from
-   within Julia, due to inlining and more agressive architecture-specific optimizations.
+   within Julia, due to inlining and more aggressive architecture-specific optimizations.
 */
 uint64_t jl_genrandom(uint64_t rngState[4]) JL_NOTSAFEPOINT
 {
@@ -798,10 +933,14 @@ JL_DLLEXPORT jl_task_t *jl_new_task(jl_function_t *start, jl_value_t *completion
     t->gcstack = NULL;
     t->excstack = NULL;
     t->started = 0;
-    t->prio = -1;
+    t->priority = 0;
     jl_atomic_store_relaxed(&t->tid, t->copy_stack ? jl_atomic_load_relaxed(&ct->tid) : -1); // copy_stacks are always pinned since they can't be moved
+    t->threadpoolid = ct->threadpoolid;
     t->ptls = NULL;
     t->world_age = ct->world_age;
+    t->reentrant_timing = 0;
+    t->reentrant_inference = 0;
+    t->inference_start_time = 0;
 
 #ifdef COPY_STACKS
     if (!t->copy_stack) {
@@ -818,6 +957,9 @@ JL_DLLEXPORT jl_task_t *jl_new_task(jl_function_t *start, jl_value_t *completion
 #endif
 #ifdef _COMPILER_TSAN_ENABLED_
     t->ctx.tsan_state = __tsan_create_fiber(0);
+#endif
+#ifdef _COMPILER_ASAN_ENABLED_
+    t->ctx.asan_fake_stack = NULL;
 #endif
     return t;
 }
@@ -829,6 +971,7 @@ JL_DLLEXPORT jl_task_t *jl_get_current_task(void)
     return pgcstack == NULL ? NULL : container_of(pgcstack, jl_task_t, gcstack);
 }
 
+
 #ifdef JL_HAVE_ASYNCIFY
 JL_DLLEXPORT jl_ucontext_t *task_ctx_ptr(jl_task_t *t)
 {
@@ -891,14 +1034,32 @@ void jl_init_tasks(void) JL_GC_DISABLED
 #endif
 }
 
-STATIC_OR_JS void NOINLINE JL_NORETURN start_task(void)
+#if defined(_COMPILER_ASAN_ENABLED_)
+STATIC_OR_JS void NOINLINE JL_NORETURN _start_task(void);
+#endif
+
+STATIC_OR_JS void NOINLINE JL_NORETURN JL_NO_ASAN start_task(void)
 {
 CFI_NORETURN
+#if defined(_COMPILER_ASAN_ENABLED_)
+    // First complete the fiber switch, otherwise ASAN will be confused
+    // when it unpoisons the stack in _start_task
+#ifdef __clang_gcanalyzer__
+    jl_task_t *ct = jl_get_current_task();
+#else
+    jl_task_t *ct = jl_current_task;
+#endif
+    jl_ptls_t ptls = ct->ptls;
+    sanitizer_finish_switch_fiber(ptls->previous_task, ct);
+    _start_task();
+}
+STATIC_OR_JS void NOINLINE JL_NORETURN _start_task(void)
+{
+CFI_NORETURN
+#endif
     // this runs the first time we switch to a task
-    sanitizer_finish_switch_fiber();
 #ifdef __clang_gcanalyzer__
     jl_task_t *ct = jl_get_current_task();
-    JL_GC_PROMISE_ROOTED(ct);
 #else
     jl_task_t *ct = jl_current_task;
 #endif
@@ -1112,7 +1273,7 @@ static void jl_start_fiber_swap(jl_ucontext_t *lastt, jl_ucontext_t *t)
 #endif
 
 #if defined(JL_HAVE_ASM)
-static void jl_start_fiber_swap(jl_ucontext_t *lastt, jl_ucontext_t *t)
+JL_NO_ASAN static void jl_start_fiber_swap(jl_ucontext_t *lastt, jl_ucontext_t *t)
 {
     assert(lastt);
 #ifdef JL_HAVE_UNW_CONTEXT
@@ -1129,7 +1290,7 @@ static void jl_start_fiber_swap(jl_ucontext_t *lastt, jl_ucontext_t *t)
     tsan_switch_to_ctx(t);
     jl_start_fiber_set(t); // doesn't return
 }
-static void jl_start_fiber_set(jl_ucontext_t *t)
+JL_NO_ASAN static void jl_start_fiber_set(jl_ucontext_t *t)
 {
     char *stk = ((char**)&t->ctx)[0];
     size_t ssize = ((size_t*)&t->ctx)[1];
@@ -1225,9 +1386,9 @@ static char *jl_alloc_fiber(_jl_ucontext_t *t, size_t *ssize, jl_task_t *owner)
     _jl_ucontext_t base_ctx;
     memcpy(&base_ctx, &ptls->base_ctx, sizeof(base_ctx));
     sigfillset(&set);
-    if (sigprocmask(SIG_BLOCK, &set, &oset) != 0) {
+    if (pthread_sigmask(SIG_BLOCK, &set, &oset) != 0) {
        jl_free_stack(stk, *ssize);
-       jl_error("sigprocmask failed");
+       jl_error("pthread_sigmask failed");
     }
     uc_stack.ss_sp = stk;
     uc_stack.ss_size = *ssize;
@@ -1259,9 +1420,9 @@ static char *jl_alloc_fiber(_jl_ucontext_t *t, size_t *ssize, jl_task_t *owner)
        jl_free_stack(stk, *ssize);
        jl_error("sigaltstack failed");
     }
-    if (sigprocmask(SIG_SETMASK, &oset, NULL) != 0) {
+    if (pthread_sigmask(SIG_SETMASK, &oset, NULL) != 0) {
        jl_free_stack(stk, *ssize);
-       jl_error("sigprocmask failed");
+       jl_error("pthread_sigmask failed");
     }
     if (&ptls->base_ctx != t) {
         memcpy(&t, &ptls->base_ctx, sizeof(base_ctx));
@@ -1361,9 +1522,13 @@ jl_task_t *jl_init_root_task(jl_ptls_t ptls, void *stack_lo, void *stack_hi)
     ct->gcstack = NULL;
     ct->excstack = NULL;
     jl_atomic_store_relaxed(&ct->tid, ptls->tid);
+    ct->threadpoolid = jl_threadpoolid(ptls->tid);
     ct->sticky = 1;
     ct->ptls = ptls;
     ct->world_age = 1; // OK to run Julia code on this task
+    ct->reentrant_timing = 0;
+    ct->reentrant_inference = 0;
+    ct->inference_start_time = 0;
     ptls->root_task = ct;
     jl_atomic_store_relaxed(&ptls->current_task, ct);
     JL_GC_PROMISE_ROOTED(ct);
@@ -1373,6 +1538,9 @@ jl_task_t *jl_init_root_task(jl_ptls_t ptls, void *stack_lo, void *stack_hi)
 #ifdef _COMPILER_TSAN_ENABLED_
     ct->ctx.tsan_state = __tsan_get_current_fiber();
 #endif
+#ifdef _COMPILER_ASAN_ENABLED_
+    ct->ctx.asan_fake_stack = NULL;
+#endif
 
 #ifdef COPY_STACKS
     // initialize the base_ctx from which all future copy_stacks will be copies
@@ -1387,13 +1555,20 @@ jl_task_t *jl_init_root_task(jl_ptls_t ptls, void *stack_lo, void *stack_hi)
 #endif
         if (jl_setjmp(ptls->copy_stack_ctx.uc_mcontext, 0))
             start_task(); // sanitizer_finish_switch_fiber is part of start_task
-        return ct;
     }
-    ssize = JL_STACK_SIZE;
-    char *stkbuf = jl_alloc_fiber(&ptls->base_ctx, &ssize, NULL);
-    ptls->stackbase = stkbuf + ssize;
-    ptls->stacksize = ssize;
+    else {
+        ssize = JL_STACK_SIZE;
+        char *stkbuf = jl_alloc_fiber(&ptls->base_ctx, &ssize, NULL);
+        if (stkbuf != NULL) {
+            ptls->stackbase = stkbuf + ssize;
+            ptls->stacksize = ssize;
+        }
+    }
 #endif
+
+    if (jl_options.handle_signals == JL_OPTIONS_HANDLE_SIGNALS_ON)
+        jl_install_thread_signal_handler(ptls);
+
     return ct;
 }
 
@@ -1407,6 +1582,11 @@ JL_DLLEXPORT int16_t jl_get_task_tid(jl_task_t *t) JL_NOTSAFEPOINT
     return jl_atomic_load_relaxed(&t->tid);
 }
 
+JL_DLLEXPORT int8_t jl_get_task_threadpoolid(jl_task_t *t)
+{
+    return t->threadpoolid;
+}
+
 
 #ifdef _OS_WINDOWS_
 #if defined(_CPU_X86_)
